type point = int * int ;;

type ensemble = point array ;;

type ordre = PLUS_PETIT | PLUS_GRAND | EGAUX

let echanger ( vecteur : 'a array ) ( i : int ) ( j : int ) : unit =
  let tampon = vecteur.(i) in
  vecteur.(i) <- vecteur.(j) ;
  vecteur.(j) <- tampon ;;

let partition
    ( vecteur : 'a array )
    ( debut : int )
    ( fin : int )
    ( ordre : 'a -> 'a -> ordre )
  : int =
  let pivot = vecteur.(debut) in
  let limite = ref debut in
  for i = debut + 1 to fin do
    if ordre vecteur.(i) pivot == PLUS_PETIT then (
     limite := !limite + 1 ;
     echanger vecteur i !limite ;
    )
  done ;
  echanger vecteur debut !limite ;
  !limite ;;

let tri_rapide
      ( vecteur : 'a array ) ( comparaison : 'a -> 'a -> ordre ) : unit =
  let rec tri_rapide_en_place ( debut : int ) ( fin : int ) : unit =
    if debut < fin - 1 then (
      let limite = partition vecteur debut fin comparaison in
      tri_rapide_en_place debut limite ;
      tri_rapide_en_place ( limite + 1 ) fin ;
    ) in
  tri_rapide_en_place 0 ( ( Array.length vecteur ) - 1 ) ;;

(*
  Paramètre = point_g, point_d : deux points de N^2
  Renvoie le résultat de la comparaison entre point_g et point_d selon l'ordre
  lexicographique.
*)
let ordre_lex ( point_g : point ) ( point_d : point ) : ordre =
  let xg, yg = point_g in
  let xd, yd = point_d in
  (*
    Paramtres n, m : deux entiers
    Renvoie le résultat de la comparaison de n et m dans le type ordre.
  *)
  let ordre_int ( n : int ) ( m :int ) : ordre =
    if n < m then PLUS_PETIT
    else if n > m then PLUS_GRAND
    else EGAUX in
  let ordre_x = ordre_int xg xd in
  match ordre_x with
  | EGAUX -> ordre_int yg yd
  | _ -> ordre_x ;;

(*
  Paramètre = ensemble : un ensemble non vide de N^2
  Renvoie le plus grand indice où apparaît le plus grand point de l'ensemble.
  Pour l'ordre lexicographique.
*)
let max_ensemble ( ensemble : ensemble ) : int =
  let ipoint_max = ref 0 in
  let cardinal = Array.length ensemble in
  for i = 1 to cardinal - 1 do
    if ordre_lex ensemble.(i) ensemble.(!ipoint_max) == PLUS_GRAND
    then ipoint_max := i
  done ;
  !ipoint_max ;;

type sens = TRIGO | ANTI_TRIGO | ALIGNE ;;

(*
  Paramètre = point_1, point_2 : deux points
  Paramètre = point : un troisième point.
  Renvoie le sens de rangement de point_1, point_2 et point.
*)

let direction
    ( point_1 : point )
    ( point_2 : point )
    ( point : point )
  : sens =
  let x1, y1 = point_1 in
  let x2, y2 = point_2 in
  let x, y = point in
  let z_produit = ( x1 - x2 ) * ( y - y2 ) - ( y1 - y2 ) * ( x - x2 ) in
  if z_produit < 0 then TRIGO
  else if z_produit > 0 then ANTI_TRIGO
  else ALIGNE ;;

let cardinal_enveloppe ( enveloppe : ensemble ) : int =
  ( Array.length enveloppe ) - 1 ;;

let rec tangente_haute
    ( enveloppe_g : ensemble )
    ( enveloppe_d : ensemble )
    ( i_point_g : int )
    ( i_point_d : int )
    : int * int =
  let point_g = enveloppe_g.( i_point_g ) in
  let point_d = enveloppe_d.( i_point_d ) in
  (* Etape (1) *)
  if i_point_g < ( cardinal_enveloppe enveloppe_g )
     && ( direction point_d point_g enveloppe_g.( i_point_g + 1 )
          == ANTI_TRIGO )
  then tangente_haute enveloppe_g enveloppe_d ( i_point_g + 1 ) ( i_point_d )
  (* Etape (2) *)
  else if i_point_d > 0
     && ( direction point_g point_d enveloppe_d.( i_point_d - 1 ) == TRIGO )
  then tangente_haute enveloppe_g enveloppe_d ( i_point_g ) ( i_point_d - 1 )
  (* Condition de terminaison attente *)
  else ( i_point_g, i_point_d ) ;;

let rec tangente_basse
    ( enveloppe_g : ensemble )
    ( enveloppe_d : ensemble )
    ( i_point_g : int )
    ( i_point_d : int )
    : int * int =
  let point_g = enveloppe_g.( i_point_g ) in
  let point_d = enveloppe_d.( i_point_d ) in
  (* Etape (3) *)
  if i_point_g > 0
     && ( direction point_d point_g enveloppe_g.( i_point_g - 1 ) == TRIGO )
  then tangente_basse enveloppe_g enveloppe_d ( i_point_g - 1 ) ( i_point_d )
  (* Etape (4) *)
  else if i_point_d < ( cardinal_enveloppe enveloppe_d )
     && ( direction point_g point_d enveloppe_d.( i_point_d + 1 )
          == ANTI_TRIGO )
  then tangente_basse enveloppe_g enveloppe_d ( i_point_g ) ( i_point_d + 1 )
  (* Condition de terminaison attente *)
  else ( i_point_g, i_point_d ) ;;

let fusion_enveloppes
      ( enveloppe_g : ensemble ) ( enveloppe_d : ensemble ) : ensemble =
  let i_point_max_g = max_ensemble enveloppe_g in
  let cardinal_enveloppe_d = cardinal_enveloppe enveloppe_d in
  let cardinal_enveloppe_g = cardinal_enveloppe enveloppe_g in
  (* On fait l'initialisation comme décrite dans l'algorithme *)
  let i_point_hg, i_point_hd =
    tangente_haute enveloppe_g enveloppe_d i_point_max_g cardinal_enveloppe_d in
  let i_point_bg, i_point_bd =
    tangente_basse enveloppe_g enveloppe_d i_point_max_g 0 in

  (*** Cas dégénérés : ***)
  let i_point_hg =
    if i_point_bg == i_point_hg then i_point_hg + 1 else i_point_hg in
  let i_point_hd =
    if i_point_hd == cardinal_enveloppe_d && i_point_bd == 0
    then i_point_hd - 1 else i_point_hd in
  (*** Construction de l'enveloppe ***)
  let cardinal_enveloppe = cardinal_enveloppe_g + i_point_bg - i_point_hg + 2
                          + i_point_hd - i_point_bd + 1 in
  let enveloppe = Array.make cardinal_enveloppe ( 0, 0 ) in
  (*** Remplissage de l'enveloppe ***)
  let indice_enveloppe = ref 0 in
  for i = 0 to i_point_bg do
      enveloppe.( !indice_enveloppe ) <- enveloppe_g.(i) ;
      indice_enveloppe := !indice_enveloppe + 1 ;
  done ;
  for i = i_point_bd to i_point_hd do
      enveloppe.( !indice_enveloppe ) <- enveloppe_d.(i) ;
      indice_enveloppe := !indice_enveloppe + 1 ;
  done ;
  for i = i_point_hg to cardinal_enveloppe_g do
      enveloppe.( !indice_enveloppe ) <- enveloppe_g.(i) ;
      indice_enveloppe := !indice_enveloppe + 1 ;
  done ;
  enveloppe ;;

(*
  Paramètre = ensemble : un sous-ensemble de N^2
  Renvoie l'enveloppe convexe de l'ensemble comme décrite dans l'énoncé
*)
let enveloppe_convexe ( ensemble : ensemble ) : ensemble =
  (* On trie l'ensemble pour effectuer les divisions succesive *)
  tri_rapide ensemble ordre_lex ;
  (* On gagne de l'espace en réalisant les divisions en place *)
  let rec enveloppe_avec_division_en_place
            ( ensemble : ensemble ) ( debut : int ) ( fin : int ) : ensemble =
    if debut == fin (* Enveloppe convexe d'un singleton *)
    then [| ensemble.(debut) ; ensemble.(debut) |]
    else ( (* Fusion des enveloppes du sous-ensemble droit et gauche *)
      fusion_enveloppes
        ( enveloppe_avec_division_en_place
            ensemble debut ( ( debut + fin ) / 2 ) )
        ( enveloppe_avec_division_en_place
            ensemble ( ( debut + fin ) / 2 + 1 ) fin )
    ) in
  enveloppe_avec_division_en_place
    ensemble 0 ( ( Array.length ensemble ) - 1 ) ;;
